#! /bin/sh
#
# Product setup script - Loki Entertainment Software

# Setup logging
if [ -n "$CX_LOG" ]
then
    [ "$CX_LOG" = "-" ] || exec 2>>"$CX_LOG"
    echo >&2
    echo "***** `date`" >&2
    echo "Starting: $0 $@" >&2
    set -x
fi

# Go to the proper setup directory (if not already there)
cd `dirname $0`

# Return the appropriate architecture string
DetectARCH()
{
	status=1
	case `uname -m` in
		i?86)  echo "x86"
			status=0;;
		x86_64)  echo "x86"
			status=0;;
		*)     echo "`uname -m`"
			status=0;;
	esac
	return $status
}

# Return all the supported versions
DetectLIBC()
{
      status=1
      if [ `uname -s` != Linux ]; then
          echo "unknown"
          return $status
      fi

      if [ -d "/lib32" ]; then
          libdir="/lib32"
      elif [ -d "/emul/ia32-linux/lib" ]; then
          libdir="/emul/ia32-linux/lib"
      else
          libdir="/lib"
      fi
      for lc in "$libdir"/libc.so.6 "$libdir"/libc.so.6.? "$libdir"/libc.so.6.*
      do
          if [ -f "$lc" ]; then
              break
          fi
      done
      
      if [ -f "$lc" ]; then
          # We cannot use strings because many systems don't have it!
          libc=""
          # Put the most common major first
          for major in 2 0 1 3 4 5 6 7 8 9
          do
              if grep "GLIBC_$major[.][0-9]" "$libdir"/libc.so.6* 2>&1 >/dev/null
              then
                  for minor in 0 1 2 3 4 5 6 7 8 9
                  do
                      if grep "GLIBC_$major[.]$minor" "$libdir"/libc.so.6* 2>&1 >/dev/null
                      then
                          libc="glibc-$major.$minor $libc"
                      fi
                  done
                  if [ -n "$libc" ]
                  then
                      # We assume only one major is supported
                      break
                  fi
              fi
          done
          if [ -z "$libc" ]; then
              echo "unknown"
          else
              echo "$libc"
              status=0
          fi
      elif [ -f "$libdir"/libc.so.5 ]; then
	      echo "libc5"
	      status=0
      else
	      echo "unknown"
      fi
      return $status
}

# Detect the Linux environment
arch=`DetectARCH`
libc=`DetectLIBC`
os=`uname -s`

# Find the installation program
try_run()
{
    setup=$1
    shift
    fatal=$1
    if [ "$1" != "" ]; then
        shift
    fi

    # First find the binary we want to run
    failed=0
    setup_bin=""
    for ver in $libc
    do
        setup_bin="setup.data/bin/$os/$arch/$ver/$setup"
        if [ -f "$setup_bin" ]
        then
            break
        fi
    done
    if [ ! -f "$setup_bin" ]; then
        setup_bin="setup.data/bin/$os/$arch/$setup"
        if [ ! -f "$setup_bin" ]; then
            failed=1
        fi
    fi
    if [ "$failed" -eq 1 ]; then
        if [ "$fatal" != "" ]; then
            libc=`echo $libc | cut -d' ' -f1`
            cat <<__EOF__
This installation does not support $libc on $os / $arch

This product requires Linux running on a pentium-compatible (i386) system.

If you have additional questions, please visit http://support.codeweavers.com
__EOF__
            exit 1
        fi
        return $failed
    fi

    # Check that $HOME belongs to us before copying the executable there
    # and before letting Loki write its stuff there
    if [ -z "$HOME" ]
    then
        cat <<__EOF__
\$HOME is not set. Point it to your home directory and try again.
__EOF__
        exit 1
    fi
    uid=`id | sed -e 's/^uid=\([0-9]*\)(.*/\1/'`
    owner=`ls -lnd "$HOME" | sed -e 's/^[^ ]*  *[^ ]*  *\([0-9]*\)  *.*$/\1/'`
    if [ "$owner" != "$uid" ]
    then
        cat <<__EOF__
The '$HOME' directory does not belong to you.
Point \$HOME to your home directory and try again.
__EOF__
        exit 1
    fi

    # Try to run the binary
    # The executable is here but we cannot execute it from CD
    setup="$HOME/.setup$$"
    cp "$setup_bin" "$setup"
    chmod 700 "$setup"
    if [ "$fatal" != "" ]; then
        if [ -n "$CX_LOG" ]; then
            "$setup" $*
            failed=$?
        else
            rcfile="$HOME/.setup$$.rc"
            (
                "$setup" $* 
                echo $? >"$rcfile"
            ) 2>&1 | egrep -v "^G[dt]k-WARNING.*(Missing charsets in FontSet creation|JISX0201\\.1976-0|JISX0208\\.1983-0|KSC5601\\.1987-0|GB2312\\.1980-0|Unable to locate loadable module in module_path:.*libqtpixmap.so)"
            failed=`cat "$rcfile"`
            rm "$rcfile"
        fi
    else
        "$setup" $* 2>/dev/null
        failed=$?
    fi
    rm -f "$setup"
    return $failed
}


# Try to run the setup program
status=0
rm -f "$setup"
try_run setup.gtk $* -fatal || {
    firstlibc=`echo $libc | cut -d' ' -f1`
    echo "The setup program seems to have failed on $arch/$firstlibc"
    echo "Check the system requirements at:"
    echo "http://www.codeweavers.com/products/cxoffice/requirements/"
    if [ `uname -m` = "x86_64" ]
    then
        echo
        echo "You might be missing the 32bit compatibility libraries"
    fi
    status=1
}
exit $status
