#! /bin/sh
#
# makeself 1.5.4
#
# $Id: makeself.sh,v 1.3 2002/07/12 22:11:00 fgouget Exp $
#
# Utility to create self-extracting tar.gz archives.
# The resulting archive is a file holding the tar.gz archive with
# a small Shell script stub that uncompresses the archive to a temporary
# directory and then executes a given script from withing that directory.
#
# Makeself home page: http://www.lokigames.com/~megastep/makeself/
#
# Version history :
# - 1.0 : Initial public release
# - 1.1 : The archive can be passed parameters that will be passed on to
#         the embedded script, thanks to John C. Quillan
# - 1.2 : Package distribution, bzip2 compression, more command line options,
#         support for non-temporary archives. Ideas thanks to Francois Petitjean
# - 1.3 : More patches from Bjarni R. Einarsson and Francois Petitjean:
#         Support for no compression (--nocomp), script is no longer mandatory,
#         automatic launch in an xterm, optional verbose output, and -target 
#         archive option to indicate where to extract the files.
# - 1.4 : Improved UNIX compatibility (Francois Petitjean)
#         Automatic integrity checking, support of LSM files (Francois Petitjean)
# - 1.5 : Many bugfixes. Optionally disable xterm spawning.
# - 1.5.1 : More bugfixes, added archive options -list and -check.
# - 1.5.2 : Cosmetic changes to inform the user of what's going on with big 
#           archives (Quake III demo)
# - 1.5.3 : Check for validity of the DISPLAY variable before launching an xterm.
#           More verbosity in xterms and check for embedded command's return value.
#           Bugfix for Debian 2.0 systems that have a different "print" command.
# - 1.5.4 : Many bugfixes. Print out a message if the extraction failed.
# - 1.5.5 : More bugfixes. Added support for SETUP_NOCHECK environment variable to
#           bypass checksum verification of archives.
#
# (C) 1998-2001 by Stphane Peter <megastep@lokigames.com>
#
# This software is released under the terms of the GNU GPL
# Please read the license at http://www.gnu.org/copyleft/gpl.html
#
VERSION=1.5.5
GZIP_CMD="gzip -c9"
GUNZIP_CMD="gzip -cd"
KEEP=n
CURRENT=n
NOX11=n
COMPRESS=gzip
TAR_ARGS=cvf
if [ "$1" = --version ]; then
	echo Makeself version $VERSION
	exit 0
fi
if [ "$1" = --bzip2 ]; then
	if which bzip2 2>&1 > /dev/null; then
		GZIP_CMD="bzip2 -9"
		GUNZIP_CMD="bzip2 -d"
		COMPRESS=bzip2
		shift 1
	else
		echo Unable to locate the bzip2 program in your \$PATH.>&2
		exit 1
	fi
else
	if [ "$1" = --nocomp ]; then
		GZIP_CMD=cat; GUNZIP_CMD=cat; COMPRESS=none
		shift 1
	fi
fi
if [ "$1" = --notemp ]; then
	KEEP=y
	shift 1
	if [ "$1" = --current ]; then
		CURRENT=y
		shift 1
	fi
fi
if [ "$1" = --nox11 ]; then
	NOX11=y
	shift 1
fi
if [ "$1" = --follow ]; then
	TAR_ARGS=cvfh
	shift 1
fi
skip=214
if [ x"$1" = x--lsm -o x"$1" = x-lsm ]; then
	shift 1
   lsm_file=$1
   [ -r $lsm_file ] && {
     l_supp=`cat $lsm_file | wc -l`
} ||  {
   echo "can't read LSM file " $lsm_file ;
  lsm_file="no_LSM";
  nl_supp=1; }
	shift 1
else
  lsm_file="no_LSM"
  nl_supp=1
fi
skip=`expr $skip + $nl_supp`
if [ "$KEEP" = n -a $# = 3 ]; then
	echo "Making a temporary archive with no embedded command does not make sense!"
	echo
	shift 1 # To force the command usage
fi
if [ $# -lt 3 ]; then
    echo "Usage: $0 [params] archive_dir file_name label [startup_script] [args]"
	echo "params can be one of those :"
	echo "    --version  : Print out Makeself version number and exit"
	echo "    --bzip2    : Compress using bzip2 instead of gzip"
	echo "    --nocomp   : Do not compress the data"
	echo "    --notemp   : The archive will create archive_dir in the"
	echo "                 current directory and uncompress in ./archive_dir"
	echo "    --current  : Used with --notemp, files will be extracted to the"
	echo "                 current directory."
    echo "    --follow   : Follow the symlinks in the archive"
	echo "    --nox11    : Disable automatic spawn of a xterm"
	echo "    --nowait   : Do not wait for user input after executing embedded program from an xterm"
	echo "    --lsm file : LSM file describing the package"
    echo Do not forget to give a fully qualified startup script name
    echo "(i.e. with a ./ prefix if inside the archive)."
    exit 1
fi

archdir=$1
archname=$2
tmpfile="${TMPDIR:=/tmp}/mkself$$"
# We don't really want to create an absolute directory...
archdirname=`basename "$1"`
USIZE=`du -ks $archdir | cut -f1`
DATE=`date`

# echo lsm=\"$lsm_contents\" >> $tmpfile

# write to the temp file first, then checksum it

echo label=\"$3\" > $tmpfile
echo script=$4 >> $tmpfile
[ x"$4" = x ] || shift 1
echo targetdir="$archdirname" >>$tmpfile
shift 3
echo scriptargs=\"$*\" >> $tmpfile
echo keep=$KEEP >> $tmpfile

cat << EODF  >> $tmpfile
TMPROOT=\${TMPDIR:=/tmp}
finish=true
xterm_loop=""
# Shells really don't like CR/LFs. So use some CR/LF-resistant
# code to perform a quick sanity check.
crlf_sentry=
if [ -n "\$crlf_sentry" ]; then #
  echo "Line ends have been converted to CR/LF." #
  echo  #
  echo "You need to download the installer again, as it is corrupt." #
  echo  #
  echo "If your download appeared to be successful, try a different" #
  echo "browser. Some browsers, such as Netscape, are known to corrupt" #
  echo "some types of binary files by trying treating them as text." #
  echo  #
  exit 2 #
fi #
LOKISCRIPTPATH=\`which \$0 2>/dev/null\`
if [ ! -f "\$LOKISCRIPTPATH" ]; then
  LOKISCRIPTPATH=\$0
fi
echo \$LOKISCRIPTPATH | egrep "^/" >/dev/null
rc=\$?
if [ \$rc -ne 0 -a -f "\$LOKISCRIPTPATH" -a -f "\`pwd\`/\$LOKISCRIPTPATH" ]; then
  LOKISCRIPTPATH="\`pwd\`/\$LOKISCRIPTPATH"
fi
export LOKISCRIPTPATH
if tail -n 1 /dev/null >/dev/null 2>&1
then
  TAIL_CMD="tail -n"
else
  TAIL_CMD="tail"
fi
[ x"\$1" = x-xwin ] && {
 finish="echo Press Return to close this window...; read junk"; xterm_loop=1; shift 1;
}
if [ x"\$1" = "x-help" -o x"\$1" = "x--help" ]; then
  cat << tac
 1) Getting help or info about \$0 :
  \$0 --help   Print this message
  \$0 --info   Print embedded info : title, default target directory, embedded script ...
  \$0 --lsm    Print embedded lsm entry (or no LSM)
  \$0 --list   Print the list of files in the archive
  \$0 --check  Checks integrity of the archive
 
 2) Running \$0 :
  \$0 [options] [additional arguments to embedded script]
  with following options (in that order)
  --confirm             Ask before running embedded script
  --keep                Do not erase target directory after running embedded script
  --target NewDirectory Extract in NewDirectory
tac
  exit 0;
fi
if [ x"\$1" = "x-lsm" -o x"\$1" = "x--lsm" ]; then
  cat << EOF_LSM
EODF
 if [ x"$lsm_file" = "xno_LSM" ]; then
    echo "no LSM" >> $tmpfile
 else
    cat $lsm_file >> $tmpfile
 fi
cat << EOF >> $tmpfile
EOF_LSM
  exit 0;
fi
if [ "\$1" = "--info" ]; then
	echo Identification: \$label
	echo Target directory: \$targetdir
	echo Uncompressed size: $USIZE KB
	echo Compression: $COMPRESS
	echo Date of packaging: $DATE
	echo script run after extraction: \$script \$scriptargs
	[ x"\$keep" = xy ] && echo "directory \$targetdir is permanent" || echo "\$targetdir will be removed after extraction"
	exit 0;
fi
if [ "\$1" = "--list" ]; then
	echo Target directory: \$targetdir
	\$TAIL_CMD +\$skip "\$0" | $GUNZIP_CMD | tar tvf -
	exit 0;
fi
if [ "\$1" = "--check" ]; then
sum1=\`\$TAIL_CMD +6 "\$0" | cksum | sed -e 's/ /Z/' -e 's/	/Z/' | cut -dZ -f1\`
[ "\$sum1" != "\$CRCsum" ] && {
  echo "Error in checksums \$sum1 != \$CRCsum"
  exit 2;
}
if [ \$MD5 != "00000000000000000000000000000000" ]; then
# space separated list of directories
  [ x"\$GUESS_MD5_PATH" = "x" ] && GUESS_MD5_PATH="/usr/local/ssl/bin"
  MD5_PATH=""
  for a in \$GUESS_MD5_PATH; do
    if which \$a/md5 >/dev/null 2>&1 ; then
       MD5_PATH=\$a;
    fi
  done
  if [ -x \$MD5_PATH/md5 ]; then
    md5sum=\`\$TAIL_CMD +6 "\$0" | \$MD5_PATH/md5\`;
    [ \$md5sum != \$MD5 ] && {
      echo Error in md5 sums \$md5sum \$MD5
      exit 2
    } || { echo check sums and md5 sums are ok; exit 0; }
  fi
  if [ ! -x \$MD5_PATH/md5 ]; then
      echo an embedded md5 sum of the archive exists but no md5 program was found in \$GUESS_MD5_PATH
      echo if you have md5 on your system, you should try :
      echo env GUESS_MD5_PATH=\"FirstDirectory SecondDirectory ...\" \$0 -check
  fi
else
  echo check sums are OK ; echo \$0 does not contain embedded md5 sum ;
fi
	exit 0;
fi
EOF

cat << EOF >> $tmpfile
[ x"\$finish" = x ] && finish=true
parsing=yes
x11=y
while [ x"\$parsing" != x ]; do
    case "\$1" in
      -confirm|--confirm) verbose=y; shift;;
      -keep|--keep) keep=y; shift;;
      -nox11|--nox11) x11=n; shift;;
      -target|--target) if [ x"\$2" != x ]; then targetdir="\$2"; keep=y; shift 2; fi;;
      *) parsing="";;
    esac
done
EOF

if [ $NOX11 = n ]; then
cat << EOF >> $tmpfile
if [ "\$x11" = "y" ]; then
    if ! tty -s; then                 # Do we have a terminal?
        if [ x"\$DISPLAY" != x -a x"\$xterm_loop" = x ]; then  # No, but do we have X?
            if xset q > /dev/null 2>&1; then # Check for valid DISPLAY variable
                GUESS_XTERMS="xterm dtterm eterm Eterm rxvt kvt konsole aterm"
                for a in \$GUESS_XTERMS; do
                    if which \$a >/dev/null 2>&1; then
                        XTERM=\$a
                        break
                    fi
                done
                chmod a+x "\$0" || echo Please add execution rights on \$0
                if [ \`echo "\$0" | cut -c1\` = / ]; then # Spawn a terminal!
                    exec \$XTERM -title "\$label" -e "\$0" -xwin "\$@"
                else
                    exec \$XTERM -title "\$label" -e "./\$0" -xwin "\$@"
                fi
            fi
        fi
    fi
fi
EOF
fi

if [ $CURRENT = n ]; then
cat << EOF >> $tmpfile
mask=\`umask\`
umask 077
if [ x"\$keep" = xy ]; then echo "Creating directory \$targetdir"; tmpdir=\$targetdir;
else tmpdir="\$TMPROOT/selfgz\$\$"; fi
mkdir \$tmpdir || {
	echo 'Cannot create target directory' \$tmpdir >&2
	echo 'you should perhaps try option -target OtherDirectory' >&2
	eval \$finish; exit 1;
}

# Create a wrapper script which can be exec'd so we don't lock the CD
wrapperbase=\`basename "\$LOKISCRIPTPATH"\`-wrapper
wrapperpath="\$tmpdir/\$wrapperbase"
echo '#!/bin/sh' >"\$wrapperpath"
[ x"\$keep" = xy ] || echo "trap 'echo Signal caught, cleaning up > /dev/tty; cd \"\$TMPROOT\"; /bin/rm -rf \"\$tmpdir\"; exit 15' 1 2 15" >>"\$wrapperpath"
echo "\$script \$scriptargs \$*" >>"\$wrapperpath"
echo "cd \$TMPROOT" >>"\$wrapperpath"
if [ x"\$keep" = xy ] ; then
    echo "exec rm -f \"\$wrapperpath\"" >>"\$wrapperpath"
else
    echo "exec rm -rf \"\$tmpdir\"" >>"\$wrapperpath"
fi
chmod +rx "\$wrapperpath"
umask \$mask
EOF
else
cat << EOF >> $tmpfile
tmpdir=.
EOF
fi
cat << EOF >> $tmpfile
location="\`pwd\`"
if [ -x /usr/bin/printf ]
then
  echo_n=/usr/bin/printf
elif [ -x /usr/ucb/echo ]
then
  echo_n="/usr/ucb/echo -n"
else
  echo_n="echo -n"
fi
if [ x\$SETUP_NOCHECK != x1 ]; then
    \$echo_n "Verifying archive integrity..."
    sum1=\`\$TAIL_CMD +6 "\$0" | cksum | sed -e 's/ /Z/' -e 's/	/Z/' | cut -dZ -f1\`
    [ "\$sum1" != "\$CRCsum" ] && {
        echo "Error in check sums \$sum1 != \$CRCsum"
        echo 
        echo "You need to download the installer again, as it is corrupt."
        echo 
        echo "If your download appeared to be successful, try a different"
        echo "browser. Some browsers, such as Netscape, are known to corrupt"
        echo "some types of binary files by trying treating them as text."
        echo 
        eval \$finish; exit 2;
    }
    echo OK
fi
if [ \$MD5 != "00000000000000000000000000000000" ]; then
# space separated list of directories
  [ x\$GUESS_MD5_PATH = x ] && GUESS_MD5_PATH="/usr/local/ssl/bin"
  MD5_PATH=""
  for a in \$GUESS_MD5_PATH; do
    if which \$a/md5 >/dev/null 2>&1 ; then
       MD5_PATH=\$a;
    fi
  done
  if [ -x \$MD5_PATH/md5 ]; then
    md5sum=\`\$TAIL_CMD +6 "\$0" | \$MD5_PATH/md5\`;
    [ \$md5sum != \$MD5 ] && {
      echo "Error in md5 sums \$md5sum \$MD5"
      eval \$finish; exit 2;
    }
  fi
fi
UnTAR() { tar xvf - || { echo Extraction failed. > /dev/tty; kill -15 \$$; } ; }
echo "Uncompressing \$label"
cd \$tmpdir ; res=3
[ "\$keep" = y ] || trap 'echo Signal caught, cleaning up > /dev/tty; cd \$TMPROOT; /bin/rm -rf \$tmpdir; eval \$finish; exit 15' 1 2 15
if (cd "\$location"; \$TAIL_CMD +\$skip "\$0"; ) | $GUNZIP_CMD | UnTAR | \
 (while read a; do \$echo_n .; done; echo; ); then
	uid=\`id | sed -e 's/^uid=\([0-9]*\)(.*/\1/'\`
	gid=\`id | sed -e 's/^.* gid=\([0-9]*\)(.*/\1/'\`
	chown -Rf \$uid:\$gid .
    res=0; if [ x"\$script" != x ]; then
		if [ x"\$verbose" = xy ]; then
			echo "OK to execute: \$script \$scriptargs \$* ? [Y/n] "
			read yn
			[ x"\$yn" = x -o x"\$yn" = xy -o x"\$yn" = xY ] && { exec "./\$wrapperbase"; res=\$?; }
		else
			exec "./\$wrapperbase" ; res=\$?
		fi
		#[ \$res != 0 ] && echo "The program returned an error code (\$res)"
	fi
    [ "\$keep" = y ] || { cd \$TMPROOT; /bin/rm -rf \$tmpdir; }
else
  echo "Cannot decompress \$0"; eval \$finish; exit 1
fi
eval \$finish; exit \$res
END_OF_STUB
EOF

# Append the compressed tar data after the stub
echo Adding files to archive named \"$archname\"...
# (cd $archdir; tar cvf - *| $GZIP_CMD ) >> $tmpfile && chmod +x $tmpfile && ..
[ -d $archdir ] || { echo "$archdir: directory does not exist" ; exit 1; }
(cd "$archdir"; tar $TAR_ARGS - * | $GZIP_CMD ) >> "$tmpfile" || { echo Aborting; exit 1; }
echo
true >> "$tmpfile" >&- ; # try to close the archive
# echo Self-extractible archive \"$archname\" successfully created.
sum1=`cksum $tmpfile | sed -e 's/ /Z/' -e 's/	/Z/' | cut -dZ -f1`
# space separated list of directories
[ x"$GUESS_MD5_PATH" = "x" ] && GUESS_MD5_PATH="/usr/local/ssl/bin"
MD5_PATH=""
for a in $GUESS_MD5_PATH; do
  if which $a/md5 >/dev/null 2>&1 ; then
     MD5_PATH=$a;
  fi
done

# now generate the real archive

# The following is the shell script stub code
echo '#! /bin/sh' > $archname
if [ $NOX11 = n ]; then
	skip=`expr $skip + 21`
fi
if [ $CURRENT = n ]; then
	skip=`expr $skip + 6`
fi
echo skip=$skip >> $archname
echo \# This script was generated using Makeself $VERSION >> $archname
if [ -x $MD5_PATH/md5 ]; then
  md5sum=`$MD5_PATH/md5 $tmpfile`;
else
  md5sum=00000000000000000000000000000000
fi
echo "CRCsum=$sum1" >> $archname
echo "MD5=$md5sum" >> $archname
cat $tmpfile >> $archname

chmod +x "$archname"
rm $tmpfile
echo Self-extractible archive \"$archname\" successfully created.

